"""Conversion from Xarray to NumpyBatch"""

from .gsp import convert_gsp_to_numpy_batch
from .nwp import convert_nwp_to_numpy_batch
from .pv import convert_pv_to_numpy_batch
from .satellite import convert_satellite_to_numpy_batch
from .sensor import convert_sensor_to_numpy_batch
from .wind import convert_wind_to_numpy_batch
